import 'package:fis_common/logger/logger.dart';
import 'package:fis_i18n/i18n.dart';
import 'package:fis_jsonrpc/services/remedical.m.dart';
import 'package:fis_measure/interfaces/process/items/item_metas.dart';
import 'package:fis_measure/process/items/item_meta_convert.dart';
import 'package:fis_measure/view/mobile_view/mobile_right_panel/mobile_measure_tool.dart';
import 'package:fis_ui/index.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:get/get.dart';

class MeasureItemCategoryModel {
  /// 类别名称
  final String name;

  /// 类别下的测量项
  final List<MeasureItemModel> items;

  MeasureItemCategoryModel({
    required this.name,
    required this.items,
  });
}

class MeasureItemModel {
  /// 外显名称（中文下只显示中文、英文下只显示英文）
  final String displayName;

  /// 测量项DTO
  final ItemMetaDTO itemMeta;

  /// 类别
  final String category;

  /// 搜索关键字
  final String searchKey;

  MeasureItemModel({
    required this.displayName,
    required this.itemMeta,
    required this.category,
    required this.searchKey,
  });
}

class MobileMoreMeasureItemDialog extends FStatefulWidget {
  final List<MobileMoreMeasureItemModesModel> measureModeList;
  final String activeItemName;
  final int? initModeIndex;
  final double? initScrollOffset;
  const MobileMoreMeasureItemDialog(
      {super.key,
      required this.measureModeList,
      required this.activeItemName,
      this.initModeIndex,
      this.initScrollOffset});

  @override
  FState<MobileMoreMeasureItemDialog> createState() =>
      _MobileMoreMeasureItemDialogState();
}

/// 更多测量项弹窗返回值
class MoreMeasureBackResult {
  final int currModeIndex;
  final double currScrollOffset;
  final ItemMeta selectedMeasureItemMeta;
  MoreMeasureBackResult({
    required this.currModeIndex,
    required this.currScrollOffset,
    required this.selectedMeasureItemMeta,
  });
}

class _MobileMoreMeasureItemDialogState
    extends FState<MobileMoreMeasureItemDialog> {
  late final ScrollController scrollController;
  final searchBarController = TextEditingController();

  /// 当前模式的下标
  int currentModeIndex = 0;

  /// 当前可用的模式列表
  List<String> currentModeNameList = [];

  /// 搜索词
  String searchWord = '';

  /// 当前模式下可用的测量项
  List<MeasureItemModel> currentModeMeasureItemList = [];

  /// 过滤后的展示结果
  List<MeasureItemModel> filteredItems = [];

  @override
  FWidget build(BuildContext context) {
    return _buildMoreMeasureItemDialog();
  }

  @override
  void initState() {
    super.initState();
    // 遍历 measureModeList 获取到当前可用的模式列表，写入 currentModeNameList
    for (final measureMode in widget.measureModeList) {
      currentModeNameList.add(measureMode.displayName);
    }
    if (widget.measureModeList.isEmpty) {
      return;
    }
    currentModeMeasureItemList =
        widget.measureModeList[currentModeIndex].availableItems;
    filteredItems = currentModeMeasureItemList.toList();
    if (widget.initModeIndex != null) {
      currentModeIndex = widget.initModeIndex!;
    }
    scrollController = ScrollController(
      initialScrollOffset: widget.initScrollOffset ?? 0,
    );
  }

  FWidget _buildDialogTitle() {
    return FPadding(
      padding: const EdgeInsets.symmetric(horizontal: 10, vertical: 5),
      child: FRow(
        children: [
          FText(
            i18nBook.measure.moreMeasurementItems.t,
            style: const TextStyle(
              color: Color.fromARGB(255, 220, 220, 220),
              fontSize: 16,
            ),
          ),
          const FSizedBox(
            width: 20,
          ),
          _buildModeSelector(currentModeNameList),
          FExpanded(
            child: _buildSearchBar(),
          ),
          FInkWell(
            onTap: () => Get.back(),
            child: const FIcon(
              Icons.close,
              color: Colors.white,
            ),
          ),
        ],
      ),
    );
  }

  /// 构建每个分类
  FWidget _buildEachCategory(MeasureItemCategoryModel category) {
    final List<MeasureItemModel> measureItemList = category.items;
    final String categoryName = category.name;
    return FColumn(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        FContainer(
          margin: const EdgeInsets.symmetric(horizontal: 10, vertical: 2),
          child: FText(
            categoryName,
            style: const TextStyle(
              color: Colors.grey,
              fontSize: 14,
            ),
            textAlign: TextAlign.start,
          ),
        ),
        FContainer(
          margin: const EdgeInsets.symmetric(horizontal: 10),
          color: Colors.grey,
          height: 1,
        ),
        FGridView.count(
          physics: const NeverScrollableScrollPhysics(),
          shrinkWrap: true,
          padding: const EdgeInsets.all(10),
          crossAxisCount: i18nBook.isCurrentRussian ? 5 : 6,
          childAspectRatio: i18nBook.isCurrentRussian ? 2 : 3,
          mainAxisSpacing: 5,
          crossAxisSpacing: 5,
          children: [
            ...measureItemList.map((e) {
              return _buildMeasureListItem(
                e,
              );
            }).toList()
          ],
        ),
      ],
    );
  }

  /// 构建更多测量项列表中的测量项按钮
  FWidget _buildMeasureListItem(
    MeasureItemModel item,
  ) {
    final bool isActive = widget.activeItemName == item.itemMeta.name &&
        widget.initModeIndex == currentModeIndex;
    return FInkWell(
      onTap: () {
        if (isActive) return;
        HapticFeedback.mediumImpact();
        ItemMeta selectedItemMeta;
        try {
          selectedItemMeta = ItemMetaConverter(item.itemMeta).output();
          final MoreMeasureBackResult result = MoreMeasureBackResult(
              currModeIndex: currentModeIndex,
              currScrollOffset: scrollController.offset,
              selectedMeasureItemMeta: selectedItemMeta);
          Get.back<MoreMeasureBackResult>(result: result);
        } catch (e) {
          logger.e(
            "Item meta -[${item.displayName}] convert error; JSON-Text: ${item.itemMeta.toJson()}",
            e,
          );
        }
      },
      child: Container(
        decoration: BoxDecoration(
          border: Border.all(
              color: isActive
                  ? Colors.white
                  : const Color.fromARGB(255, 71, 71, 71)),
          borderRadius: BorderRadius.circular(4),
        ),
        child: FCenter(
          child: FText(
            item.displayName,
            textAlign: TextAlign.center,
            style: TextStyle(
              fontSize: 12,
              color: isActive ? Colors.white : Colors.grey,
            ),
          ),
        ),
      ),
    );
  }

  /// 构建模式选择按钮
  FWidget _buildModeSelector(List<String> currentModeNameList) {
    return FRow(
      mainAxisAlignment: MainAxisAlignment.center,
      children: currentModeNameList
          .asMap()
          .entries
          .map<FWidget>(
            (entry) => FInkWell(
              onTap: () => _changeCurrModeByIndex(entry.key),
              child: FContainer(
                constraints: const BoxConstraints(
                  minWidth: 40,
                ),
                margin: const EdgeInsets.symmetric(horizontal: 5),
                padding: const EdgeInsets.all(5),
                decoration: BoxDecoration(
                  border: entry.key == currentModeIndex
                      ? Border.all(color: Colors.white)
                      : Border.all(
                          color: const Color.fromRGBO(124, 124, 124, 1)),
                  borderRadius: BorderRadius.circular(5),
                ),
                child: FCenter(
                  child: FText(
                    entry.value,
                    style: TextStyle(
                      fontSize: 16,
                      color: entry.key == currentModeIndex
                          ? Colors.white
                          : Colors.grey,
                    ),
                  ),
                ),
              ),
            ),
          )
          .toList(),
    );
  }

  /// 构建更多测量项弹窗容器
  FWidget _buildMoreMeasureItemDialog() {
    return FDialog(
      backgroundColor: const Color.fromARGB(255, 36, 36, 36),
      shape: const RoundedRectangleBorder(
        borderRadius: BorderRadius.all(
          Radius.circular(7),
        ),
      ),
      clipBehavior: Clip.hardEdge,
      child: FColumn(
        mainAxisSize: MainAxisSize.min,
        children: [
          _buildDialogTitle(),
          FContainer(
            color: const Color.fromARGB(255, 36, 36, 36),
            width: Get.width,
            height: Get.height * 0.6,
            child: FContainer(
              child: _buildMoreMeasureItemList(),
            ),
          ),
          const FSizedBox(
            height: 10,
          )
        ],
      ),
    );
  }

  /// 构建更多测量项列表
  FWidget _buildMoreMeasureItemList() {
    if (currentModeMeasureItemList.isEmpty) {
      return FContainer(
        color: const Color.fromARGB(255, 36, 36, 36),
        child: FCenter(
          child: FText(
            i18nBook.measure.unsupportedMeasurementMode.t,
            style: const TextStyle(
              color: Colors.grey,
              fontSize: 16,
            ),
          ),
        ),
      );
    }
    if (filteredItems.isEmpty) {
      return FContainer(
        color: const Color.fromARGB(255, 36, 36, 36),
        child: FCenter(
          child: FText(
            i18nBook.measure.noMatchingMeasurement.t,
            style: const TextStyle(
              color: Colors.grey,
              fontSize: 16,
            ),
          ),
        ),
      );
    }

    final List<MeasureItemCategoryModel> categories =
        _getClassifiedData(filteredItems);
    return FContainer(
      color: const Color.fromARGB(255, 36, 36, 36),
      child: FScrollbar(
        controller: scrollController,
        isAlwaysShown: true,
        child: FListView(
          shrinkWrap: true,
          controller: scrollController,
          children: [
            FColumn(
              children: [
                ...categories.map((e) {
                  return _buildEachCategory(e);
                }).toList(),
              ],
            )
          ],
        ),
      ),
    );
  }

  FWidget _buildSearchBar() {
    return FContainer(
      height: 40,
      margin: const EdgeInsets.symmetric(horizontal: 10),
      decoration: BoxDecoration(
        color: const Color.fromARGB(255, 36, 36, 36),
        borderRadius: BorderRadius.circular(5),
      ),
      child: FRow(
        children: [
          const FSizedBox(
            width: 30,
          ),
          const FIcon(
            Icons.search,
            color: Colors.grey,
          ),
          const FSizedBox(
            width: 10,
          ),
          FExpanded(
            child: FTextField(
              controller: searchBarController,
              style: const TextStyle(
                color: Colors.white,
              ),
              decoration: InputDecoration(
                hintText: i18nBook.common.search.t,
                hintStyle: const TextStyle(
                  color: Colors.grey,
                ),
                border: InputBorder.none,
              ),
              onChanged: (value) {
                _filterMeasureItems(value);
              },
            ),
          ),
        ],
      ),
    );
  }

  void _changeCurrModeByIndex(int index) {
    currentModeIndex = index;
    currentModeMeasureItemList =
        widget.measureModeList[currentModeIndex].availableItems;
    _filterMeasureItems(searchWord);
  }

  void _filterMeasureItems(String value) {
    searchWord = value;
    value = value.toLowerCase();
    filteredItems.clear(); // 清除任何现有的搜索结果
    for (MeasureItemModel itemModel in currentModeMeasureItemList) {
      if (itemModel.searchKey.contains(value)) {
        filteredItems.add(itemModel);
      }
    }
    setState(() {});
  }

  List<MeasureItemCategoryModel> _getClassifiedData(
    List<MeasureItemModel> items,
  ) {
    final Map<String, List<MeasureItemModel>> categoryMap = {};
    for (MeasureItemModel item in items) {
      if (categoryMap.containsKey(item.category)) {
        categoryMap[item.category]!.add(item);
      } else {
        categoryMap[item.category] = [item];
      }
    }
    List<MeasureItemCategoryModel> classifiedData = categoryMap.entries
        .map((e) => MeasureItemCategoryModel(
              name: e.key,
              items: e.value,
            ))
        .toList();

    classifiedData.sort((a, b) => a.name.compareTo(b.name));

    /// 遍历 classifiedData 将 common 类别放在最前面
    int commonIndex = classifiedData
        .indexWhere((element) => element.name.toLowerCase().contains('common'));
    if (commonIndex != -1) {
      final MeasureItemCategoryModel commonCategory =
          classifiedData[commonIndex];
      classifiedData.removeAt(commonIndex);
      classifiedData.insert(0, commonCategory);
    }

    return classifiedData;
  }
}
